/*
 * ZGrabMain.c
 *
 * Copyright 2003 Codejedi Inc.
 * Released under the GPL. Have fun, eat cake.
 *
 */
 
#include <PalmOS.h>
#include <VfsMgr.h>
#include "ZGrab.h"
#include "ZGrabRsc.h"
#include "stdfile.h"
#include "pcx.h"
#include "bmp.h"

/*********************************************************************
 * crap
 *********************************************************************/

char *resolution_s[] = {
  "160", "320", "480", "640"
};

resolution_t g_width = res480;
resolution_t g_height = res320;

/*********************************************************************
 * Global variables
 *********************************************************************/

/* g_prefs
 * cache for application preferences during program execution */
ZGrabPreferenceType g_prefs;

/*********************************************************************
 * Internal Constants
 *********************************************************************/

/* Define the minimum OS version we support */
#define ourMinVersion    sysMakeROMVersion(3,5,0,sysROMStageDevelopment,0)
#define kPalmOS20Version sysMakeROMVersion(2,0,0,sysROMStageDevelopment,0)

/*********************************************************************
 * Internal Functions
 *********************************************************************/

/*
 * FUNCTION: GetObjectPtr
 *
 * DESCRIPTION:
 *
 * This routine returns a pointer to an object in the current form.
 *
 * PARAMETERS:
 *
 * formId
 *     id of the form to display
 *
 * RETURNED:
 *     address of object as a void pointer
 */

static void * GetObjectPtr(UInt16 objectID)
{
    FormType * frmP;

    frmP = FrmGetActiveForm();
    return FrmGetObjectPtr(frmP, FrmGetObjectIndex(frmP, objectID));
}

/*
 * FUNCTION: MainFormInit
 *
 * DESCRIPTION: This routine initializes the MainForm form.
 *
 * PARAMETERS:
 *
 * frm
 *     pointer to the MainForm form.
 */

static void MainFormInit(FormType *frmP)
{
  ListType *list;

#ifdef BMPSUPPORT
  update_field ( frmP, MainTypeField, "BMP" );
#else
  update_field ( frmP, MainTypeField, "PCX" );
#endif

  /* width
   */
  list = FrmGetObjectPtr ( frmP, FrmGetObjectIndex ( frmP, MainWidthList ) );
  LstSetListChoices ( list, resolution_s, resmax );
  LstSetSelection ( list, g_width );

  CtlSetLabel ( GetObjectPtr ( MainWidthPopTrigger ),
		resolution_s [ g_width ] );

  /* height
   */
  list = FrmGetObjectPtr ( frmP, FrmGetObjectIndex ( frmP, MainHeightList ) );
  LstSetListChoices ( list, resolution_s, resmax );
  LstSetSelection ( list, g_height );

  CtlSetLabel ( GetObjectPtr ( MainHeightPopTrigger ),
		resolution_s [ g_height ] );

  return;
}

/*
 * FUNCTION: MainFormDoCommand
 *
 * DESCRIPTION: This routine performs the menu command specified.
 *
 * PARAMETERS:
 *
 * command
 *     menu item id
 */

static Boolean MainFormDoCommand(UInt16 command)
{
    Boolean handled = false;
#if 0
    FormType * frmP;

    switch (command)
    {
    case OptionsAboutZGrab:

        /* Clear the menu status from the display */
        MenuEraseStatus(0);

        /* Display the About Box. */
        frmP = FrmInitForm (AboutForm);
        FrmDoDialog (frmP);                    
        FrmDeleteForm (frmP);

        handled = true;
        break;
    }
#endif
    
    return handled;
}

/*
 * FUNCTION: MainFormHandleEvent
 *
 * DESCRIPTION:
 *
 * This routine is the event handler for the "MainForm" of this 
 * application.
 *
 * PARAMETERS:
 *
 * eventP
 *     a pointer to an EventType structure
 *
 * RETURNED:
 *     true if the event was handled and should not be passed to
 *     FrmHandleEvent
 */

static Boolean MainFormHandleEvent ( EventType * eventP ) {
  Boolean handled = false;
  FormType * frmP = FrmGetActiveForm();

  switch ( eventP -> eType ) {

  case menuEvent:
    return MainFormDoCommand(eventP->data.menu.itemID);

  case frmOpenEvent:
    frmP = FrmGetActiveForm();
    MainFormInit ( frmP );
    FrmDrawForm ( frmP );
    handled = true;
    break;
            
  case frmUpdateEvent:
    /* To do any custom drawing here, first call
     * FrmDrawForm(), then do your drawing, and
     * then set handled to true. */
    break;

  case popSelectEvent:
    handled = true;

    switch ( eventP -> data.popSelect.controlID ) {

    case MainWidthPopTrigger:
      {
	Int16 s = -1;
	ListType *list;

	list = FrmGetObjectPtr ( frmP,
				 FrmGetObjectIndex ( frmP, MainWidthList ) );

	s = LstGetSelection ( list );

	if ( s >= 0 ) {
	  g_width = s;
	  CtlSetLabel ( GetObjectPtr ( MainWidthPopTrigger ),
			resolution_s [ g_width ] );
	}

      }
      break;

    case MainHeightPopTrigger:
      {
	Int16 s = -1;
	ListType *list;

	list = FrmGetObjectPtr ( frmP,
				 FrmGetObjectIndex ( frmP, MainHeightList ) );

	s = LstGetSelection ( list );

	if ( s >= 0 ) {
	  g_height = s;
	  CtlSetLabel ( GetObjectPtr ( MainHeightPopTrigger ),
			resolution_s [ g_height ] );
	}

      }
      break;

    } // switch

    break;

  case ctlSelectEvent:

    switch ( eventP -> data.ctlSelect.controlID ) {

    case MainOnButton:
      notify_register();
      handled = true;
      break;

    case MainOffButton:
      notify_unregister();
      handled = true;
      break;

#if 0
    case MainViewButton:
      {
	UInt8 pal [ 256 * 3 ];
	UInt32 x, y;
	UInt8 *pic;

	pic = LoadPCX ( "a.pcx", pal, &x, &y );

	{
	  UInt32 depth = 16;
	  Boolean bool = true;
	  WinScreenMode ( winScreenModeSet, NULL, NULL, &depth, &bool );
	}

#if 1
	{
	  WinHandle winh;
	  BitmapType *winb;
	  UInt16 *screenbuf;

	  UInt16 *iter = pic;
	  UInt32 limit;

	  // get screen buffer
	  winh = WinGetDisplayWindow();
	  winb = WinGetBitmap ( winh );
	  screenbuf = (UInt16*) BmpGetBits ( winb );

	  limit = x * y;
	  while ( limit ) {
	    *screenbuf++ = 37;
	    limit--;
	  }

	  screenbuf = (UInt16*) BmpGetBits ( winb );
	  iter = pic;

	  limit = x * y;
	  while ( limit ) {
	    *screenbuf++ = *iter++;
	    limit--;
	  }

	  DEBUGS ( "foo" );

	}
#endif

	MemPtrFree ( pic );

      }
      handled = true;
      break;
#endif

#if 0
    case MainSnappyButton:
      handled = true;
      do_screenshot();
      break;
#endif

    } // switch

    break;

  default:
    break;

  }
    
  return handled;
}

/*
 * FUNCTION: AppHandleEvent
 *
 * DESCRIPTION: 
 *
 * This routine loads form resources and set the event handler for
 * the form loaded.
 *
 * PARAMETERS:
 *
 * event
 *     a pointer to an EventType structure
 *
 * RETURNED:
 *     true if the event was handled and should not be passed
 *     to a higher level handler.
 */

static Boolean AppHandleEvent(EventType * eventP)
{
    UInt16 formId;
    FormType * frmP;

    if (eventP->eType == frmLoadEvent)
    {
        /* Load the form resource. */
        formId = eventP->data.frmLoad.formID;
        frmP = FrmInitForm(formId);
        FrmSetActiveForm(frmP);

        /* Set the event handler for the form.  The handler of the
         * currently active form is called by FrmHandleEvent each
         * time is receives an event. */
        switch (formId)
        {
        case MainForm:
            FrmSetEventHandler(frmP, MainFormHandleEvent);
            break;

        default:
            break;

        }
        return true;
    }

    return false;
}

/*
 * FUNCTION: AppEventLoop
 *
 * DESCRIPTION: This routine is the event loop for the application.
 */

static void AppEventLoop(void)
{
    UInt16 error;
    EventType event;

    do {
        /* change timeout if you need periodic nilEvents */
        EvtGetEvent(&event, evtWaitForever);

        if (! SysHandleEvent(&event))
        {
            if (! MenuHandleEvent(0, &event, &error))
            {
                if (! AppHandleEvent(&event))
                {
                    FrmDispatchEvent(&event);
                }
            }
        }
    } while (event.eType != appStopEvent);
}

/*
 * FUNCTION: AppStart
 *
 * DESCRIPTION:  Get the current application's preferences.
 *
 * RETURNED:
 *     errNone - if nothing went wrong
 */

static Err AppStart ( void ) {
  //UInt16 prefsSize;
  Err error;
  UInt16 voliter;
  FileRef dirRef = 0;
  //UInt32 depth;
  UInt32 width, height;

  /* create our directory if we can't open it.. */
  voliter = palm_vfs_iter();

  error = VFSFileOpen ( voliter, VFSPATH, vfsModeRead, &dirRef );

  if ( error != errNone ) {
    error = VFSDirCreate ( voliter, VFSROOT );
    error = VFSDirCreate ( voliter, VFSPATH );

    error = VFSFileOpen ( voliter, VFSPATH, vfsModeRead, &dirRef );

    if ( error != errNone ) {
      FrmCustomAlert ( MessageAlert,
		       "ZGrab requires expansion media for its "
		       "data files; put media in first slot.", NULL, NULL );
      return ( -1 );
    }

  }

  VFSFileClose ( dirRef );

#if 0
  /* set screen depth for testing..
   */
  depth = 16;
  WinScreenMode ( winScreenModeSet, NULL, NULL, &depth, NULL );
#endif
    
  /* fetch prefs
   */
  pref_fetch ( &width, &height );
  g_width = (resolution_t) width;
  g_height = (resolution_t) height;

  return errNone;
}

/*
 * FUNCTION: AppStop
 *
 * DESCRIPTION: Save the current state of the application.
 */

static void AppStop(void)
{
        
  /* write prefs
   */
  pref_write ( g_width, g_height );

  /* Close all the open forms. */
  FrmCloseAllForms();

}

/* all code from here to end of file should use no global variables */
#pragma warn_a5_access on

/*
 * FUNCTION: RomVersionCompatible
 *
 * DESCRIPTION: 
 *
 * This routine checks that a ROM version is meet your minimum 
 * requirement.
 *
 * PARAMETERS:
 *
 * requiredVersion
 *     minimum rom version required
 *     (see sysFtrNumROMVersion in SystemMgr.h for format)
 *
 * launchFlags
 *     flags that indicate if the application UI is initialized
 *     These flags are one of the parameters to your app's PilotMain
 *
 * RETURNED:
 *     error code or zero if ROM version is compatible
 */

static Err RomVersionCompatible(UInt32 requiredVersion, UInt16 launchFlags)
{
    UInt32 romVersion;

    /* See if we're on in minimum required version of the ROM or later. */
    FtrGet(sysFtrCreator, sysFtrNumROMVersion, &romVersion);
    if (romVersion < requiredVersion)
    {
        if ((launchFlags & 
            (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp)) ==
            (sysAppLaunchFlagNewGlobals | sysAppLaunchFlagUIApp))
        {
            FrmAlert (RomIncompatibleAlert);

            /* Palm OS versions before 2.0 will continuously relaunch this
             * app unless we switch to another safe one. */
            if (romVersion < kPalmOS20Version)
            {
                AppLaunchWithCommand(
                    sysFileCDefaultApp, 
                    sysAppLaunchCmdNormalLaunch, NULL);
            }
        }

        return sysErrRomIncompatible;
    }

    return errNone;
}

/*
 * FUNCTION: ZGrabPalmMain
 *
 * DESCRIPTION: This is the main entry point for the application.
 *
 * PARAMETERS:
 *
 * cmd
 *     word value specifying the launch code. 
 *
 * cmdPB
 *     pointer to a structure that is associated with the launch code
 *
 * launchFlags
 *     word value providing extra information about the launch
 *
 * RETURNED:
 *     Result of launch, errNone if all went OK
 */

static UInt32 ZGrabPalmMain(
    UInt16 cmd, 
    MemPtr /*cmdPBP*/, 
    UInt16 launchFlags)
{
    Err error;

    error = RomVersionCompatible (ourMinVersion, launchFlags);
    if (error) return (error);

    switch (cmd)
    {
    case sysAppLaunchCmdNormalLaunch:
        error = AppStart();
        if (error) 
            return error;

        /* start application by opening the main form
         * and then entering the main event loop */
        FrmGotoForm(MainForm);
        AppEventLoop();
        
        AppStop();
        break;

    default:
        break;
    }

    return errNone;
}

/*
 * FUNCTION: PilotMain
 *
 * DESCRIPTION: This is the main entry point for the application.
 * 
 * PARAMETERS:
 *
 * cmd
 *     word value specifying the launch code. 
 *
 * cmdPB
 *     pointer to a structure that is associated with the launch code
 *
 * launchFlags
 *     word value providing extra information about the launch.
 *
 * RETURNED:
 *     Result of launch, errNone if all went OK
 */
 
UInt32 PilotMain(UInt16 cmd, MemPtr cmdPBP, UInt16 launchFlags)
{
    return ZGrabPalmMain(cmd, cmdPBP, launchFlags);
}

/* turn a5 warning off to prevent it being set off by C++
 * static initializer code generation */
#pragma warn_a5_access reset

UInt16 palm_vfs_count ( void ) {
  UInt16 volref = 0;
  UInt32 voliter = vfsIteratorStart;
  UInt16 count = 0;

  if ( VFSVolumeEnumerate ( &volref,  &voliter ) != errNone ) {
    return ( 0 ); // error
  }

  count = 1;

  while ( VFSVolumeEnumerate ( &volref,  &voliter ) == errNone ) {
    //if ( voliter != vfsIteratorStop ) {
    count++;
    //}
  }

  return ( count );
}

UInt16 palm_vfs_iter ( void ) {
  UInt16 volref = 0;
  UInt32 voliter = vfsIteratorStart;
  FileRef file;
  UInt16 firstvolref;

  // can we enumerate at all?
  if ( VFSVolumeEnumerate ( &volref,  &voliter ) != errNone ) {
    return ( 0 );
  }

  // store first succeeding volume
  firstvolref = volref;

  // can we open our directory? If so, we found our directory :)
  if ( VFSFileOpen ( volref, VFSPATH, vfsModeRead, &file ) == errNone ) {
    VFSFileClose ( file );
    return ( volref );
  }

  // look through other volumes, trying to find our directory
  while ( VFSVolumeEnumerate ( &volref,  &voliter ) == errNone ) {

    if ( VFSFileOpen ( volref, VFSPATH, vfsModeRead, &file ) == errNone ) {
      VFSFileClose ( file );
      return ( volref );
    }

  }

  //return ( voliter );
  return ( firstvolref );
}

UInt8 notify_register ( void ) {
  LocalID dbid;
  Err error;
  UInt32 os400 = sysMakeROMVersion(4,0,0,sysROMStageRelease,0); /* 3.9 :/ */
  UInt32 romversion = 0;
  UInt8 palm_os_41 = 0;

  FtrGet ( sysFtrCreator, sysFtrNumROMVersion, &romversion );

  if ( romversion > os400 ) {
    palm_os_41 = 1;
    //FrmCustomAlert ( MessageAlert, "OS41", NULL, NULL );
  }

  if ( ! palm_os_41 ) {
    return ( 1 ); // fake success
  }

  dbid = DmFindDatabase ( 0, "ZGrab" );

  if ( ! dbid ) {
    DEBUGS ( "Couldn't (find) register application!" );
    return ( 0 ); // couldn't find self!
  }

  error = SysNotifyRegister ( 0 /* cardno */, dbid,
			      sysNotifySleepNotifyEvent,
			      /*sysNotifyVirtualCharHandlingEvent,*/
			      NotifyCallback /* WaveletNotifyCallback */,
			      sysNotifyNormalPriority,
			      NULL );

  if ( error ) {
    return ( 0 );
  }

  return ( 1 );
}

UInt8 notify_unregister ( void ) {
  LocalID dbid;
  Err error;
  UInt32 os400 = sysMakeROMVersion(4,0,0,sysROMStageRelease,0); /* 3.9 :/ */
  UInt32 romversion = 0;
  UInt8 palm_os_41 = 0;

  FtrGet ( sysFtrCreator, sysFtrNumROMVersion, &romversion );

  if ( romversion > os400 ) {
    palm_os_41 = 1;
  }

  if ( ! palm_os_41 ) {
    return ( 1 ); // fake success
  }

  dbid = DmFindDatabase ( 0, "ZGrab" );

  if ( ! dbid ) {
    return ( 0 ); // couldn't find self!
  }

  error = SysNotifyUnregister ( 0 /* cardno */, dbid,
				sysNotifySleepNotifyEvent,
				/*sysNotifyVirtualCharHandlingEvent,*/
				sysNotifyNormalPriority );

  if ( error ) {
    return ( 0 );
  }

  return ( 1 );
}

/* test if present or not? */
UInt8 notify_present_p ( void ) {

  /* try and remove it .. if no error, it was on, so put it back and
   * return that it was active.
   */
  if ( notify_unregister() ) {
    // successfully turned off, so we know it was active
    notify_register();
    return ( 1 ); // present
  }

  // we know it was off, since failed to deactivate
  return ( 0 );
}

Err NotifyCallback ( SysNotifyParamType *notifyParamsP ) {
#if 0
  LocalID dbid = DmFindDatabase ( 0, "ZGrab" );
  DmOpenRef themedb;

  if  ( ! dbid ) {
    SndPlaySystemSound ( sndError );
    return ( 0 );
  }

  themedb = DmOpenDatabase ( 0, dbid, dmModeReadOnly );

  if ( ! dbid ) {
    SndPlaySystemSound ( sndError );
    return ( 0 );
  }
#endif

  do_screenshot();

#if 0
  if ( FrmAlert ( AnotherAlert ) == AnotherYes ) {
    notify_register();
  } else {
    notify_unregister();
  }

  DmCloseDatabase ( themedb );
#endif

  return ( 0 /* always by RefGuide */ );
}

UInt16 find_hardkey ( void ) {
  UInt16 bits = 0;

  /* which hardkey are we on, if any? */

  if ( PrefGetPreference ( prefHard1CharAppCreator ) == CREATORID ) {
    bits = keyBitHard1;
  }

  if ( PrefGetPreference ( prefHard2CharAppCreator ) == CREATORID ) {
    bits = keyBitHard2;
  }

  if ( PrefGetPreference ( prefHard3CharAppCreator ) == CREATORID ) {
    bits = keyBitHard3;
  }

  if ( PrefGetPreference ( prefHard4CharAppCreator ) == CREATORID ) {
    bits = keyBitHard4;
  }

  return ( bits );
}

UInt8 do_screenshot ( void ) {
  WinHandle winh;
  BitmapType *winb;
  UInt8 *screenbuf;
  RGBColorType c;
  UInt16 iter;
  UInt8 palette [ 256 * 3 ];
  UInt8 retval;
  UInt32 depth;
  char filename [ 30 ];
  UInt16 timer;
  UInt32 width, height;
  UInt16 resolution_map [ 4 ];

  // map enum to real sizes
  resolution_map [ 0 ] = 160;
  resolution_map [ 1 ] = 320;
  resolution_map [ 2 ] = 480;
  resolution_map [ 3 ] = 640;

  // make palette
  for ( iter = 0; iter < 256; iter++ ) {
    WinIndexToRGB ( iter, &c );
    palette [ ( iter * 3 ) + 0 ] = c.r;
    palette [ ( iter * 3 ) + 1 ] = c.g;
    palette [ ( iter * 3 ) + 2 ] = c.b;
  }

  // get screen buffer
  winh = WinGetDisplayWindow();
  winb = WinGetBitmap ( winh );
  screenbuf = (UInt8*) BmpGetBits ( winb );

  // what depth?
  WinScreenMode ( winScreenModeGet, NULL, NULL, &depth, NULL );

#if 0
  if ( depth == 16 ) {
    WinDrawChar ( 'F', 10, 10 );
  } else if ( depth == 8 ) {
    WinDrawChar ( '8', 10, 10 );
  } else {
    WinDrawChar ( 'U', 10, 10 );
  }
#endif

  // what size?
  pref_fetch ( &width, &height );

  // determine filename
  {
    char fmt [ 20 ];
    UInt8 foo = 0;
    fmt [ foo++ ] = '%';
    fmt [ foo++ ] = 'l';
    fmt [ foo++ ] = 'u';
    fmt [ foo++ ] = '.';
#ifdef BMPSUPPORT
    fmt [ foo++ ] = 'b';
    fmt [ foo++ ] = 'm';
    fmt [ foo++ ] = 'p';
#else
    fmt [ foo++ ] = 'p';
    fmt [ foo++ ] = 'c';
    fmt [ foo++ ] = 'x';
#endif
    fmt [ foo++ ] = '\0';

    timer = TimGetSeconds() & 0xFFFF;
    StrPrintF ( filename, fmt, timer );

#ifdef BMPSUPPORT
    retval = SaveBMP ( filename, screenbuf,
		       resolution_map [ width ], resolution_map [ height ],
		       depth == 16 );
#else
    retval = SavePCX ( filename, screenbuf,
		       resolution_map [ width ], resolution_map [ height ],
		       palette, depth == 16 );
#endif

  } // fmt

  if ( retval ) {
    return ( 0 ); // fail
  }

  return ( 1 );
}

int update_field ( FormPtr f, int obid, char *newtext ) {
  MemHandle newhandle, oldhandle;
  FieldPtr field;

  newhandle = MemHandleNew ( StrLen ( newtext ) + 1 );

  if ( ! newhandle ) {
    return ( 0 );
  }

  StrCopy ( MemHandleLock ( newhandle ), newtext );
  MemHandleUnlock ( newhandle );

  field = FrmGetObjectPtr ( f, FrmGetObjectIndex ( f, obid ) );

  oldhandle = FldGetTextHandle ( field );

  FldSetTextHandle ( field, newhandle );

  if ( oldhandle ) {
    MemHandleFree ( oldhandle );
  }

  return ( 1 );
}

UInt8 pref_fetch ( UInt32 *r_width, UInt32 *r_height ) {
  pref_t prefs;
  UInt16 rsize = 0;

  if ( PrefGetAppPreferences ( CREATORID, PREFID, &prefs,
			       &rsize, true ) == noPreferenceFound )
  {
    *r_width = res480;
    *r_height = res320;
    return ( 1 );
  }

  //DEBUGU8 ( rsize );
  //DEBUGU32 ( rsize );

  if ( rsize != sizeof(prefs) ) {
    *r_width = res480;
    *r_height = res320;
    return ( 1 );
  }

  PrefGetAppPreferences ( CREATORID, PREFID, &prefs, &rsize, true );

  *r_width = prefs.width;
  *r_height = prefs.height;

  return ( 1 );
}

UInt8 pref_write ( UInt32 width, UInt32 height ) {
  pref_t prefs;

  prefs.width = width;
  prefs.height = height;
  prefs.format = 0;
  prefs.flags = 0;
  prefs.nil1 = 0;
  prefs.nil2 = 0;
  prefs.nil3 = 0;
  prefs.nil4 = 0;

  PrefSetAppPreferences ( CREATORID, PREFID, PREFVERSION, 
			  &prefs, sizeof ( prefs ), true );

  return ( 1 );
}
