#include <PalmOS.h>
#include <VfsMgr.h>
#include "zgrabrsc.h"
#include "zgrab.h"
#include "stdfile.h"

FILE *fopen ( signed char *filename, char *mode ) {
  UInt32 voliter = palm_vfs_iter();
  char path [ 256 ];
  Err error;
  FILE *myfile = MemPtrNew ( sizeof(FILE) );

#ifdef DEBUGFILE
  DEBUGS( __FUNCTION__);
#endif

  StrCopy ( myfile -> filename, filename );
  myfile -> tail = 0;
  myfile -> ref = 0;
  myfile -> size = 0;

  if ( StrChr ( mode, 'r' ) && StrChr ( mode, 'w' ) ) {
    myfile -> bufferp = 0; // no buffering on read+write files
    DEBUGS("Internal eror: Game fopen() is confusing");
  } else if ( StrChr ( mode, 'w' ) ) {
    myfile -> bufferp = 2; // buffering single direction stream
  } else {
    myfile -> bufferp = 1; // buffering single direction stream
  }

  // looks like general purpose work, so don't screw with path
  StrPrintF ( path, "%s%s", VFSPATH, filename );

  //DEBUGS(path);

  error = VFSFileOpen ( voliter, path, vfsModeReadWrite, & (myfile->ref) );

  if ( error != errNone ) {

    if ( StrChr ( mode, 'w' ) ) {
      /* couldn't open it, but try and create it.. */
      error = VFSFileCreate ( voliter, path );

      if ( error != errNone ) {
	return ( NULL );
      }

      error = VFSFileOpen ( voliter, path, vfsModeReadWrite, & (myfile->ref) );

      if ( error != errNone ) {
	MemPtrFree ( myfile );
	return ( NULL );
      }

    } else {
      MemPtrFree ( myfile );
      return ( NULL );
    }

  }

  VFSFileSize ( myfile -> ref, & (myfile -> size) );

#ifdef DEBUGFILE
  DEBUGS( "Exit");
#endif

  return ( myfile );
}

Int8 fclose ( FILE *file ) {

#ifdef DEBUGFILE
  DEBUGS(__FUNCTION__);
#endif

  if ( file -> bufferp == 2 && file -> tail ) {
    // write buffer, so we need to empty the buffer into the stream
    UInt32 result;
    VFSFileWrite ( file -> ref, file -> tail, file -> buffer, &result );
  }

  VFSFileClose ( file -> ref );

  MemPtrFree ( file );

#ifdef DEBUGFILE
  DEBUGS( "Exit");
#endif

  return ( 0 );
}

UInt32 fwrite ( char *buf, UInt32 size, UInt32 num, FILE *file ) {
  UInt32 result;

#ifdef DEBUGFILE
  DEBUGS(__FUNCTION__);
  DEBUGU32(size*num);
#endif

  // for fwrite, lets assume they're calling it since they want to write
  // a block of stuff, not just a single character; so first lets flush
  // any buffer so far, and then write out their buffer
  // (later could shove their buf into our file buffer..)
#if 0
  Err error;

  if ( file -> tail ) {
    VFSFileWrite ( file -> ref, file -> tail, file -> buffer, &result );
    file -> tail = 0;
  }

  if ( size*num ) {
    error = VFSFileWrite ( file -> ref, size*num, buf, &result );
  }
#else
  result = size*num;
  while ( result ) {
    fputc ( *buf++, file );
    result--;
  }
#endif

#ifdef DEBUGFILE
  DEBUGS( "Exit");
#endif

  //return ( result );
  return ( num );
}

int fputc ( UInt8 c, FILE *file ) {

#ifdef DEBUGFILE
  DEBUGS(__FUNCTION__);
#endif

  if ( file == stdout ) {
    //screen_push ( c );
  } else {

    //draw_progress_bar ( 100 * file -> tail / 255 );

    // buffer large enough to hold another char?
    if ( file -> tail == TAILMAX ) {
      UInt32 result;
      VFSFileWrite ( file -> ref, file -> tail, file -> buffer, &result );
      file -> tail = 0;
      //fwrite ( &c, 1, 1, file );
    }

    // insert into buffer
    file -> buffer [ file -> tail++ ] = c;

  }

#ifdef DEBUGFILE
  DEBUGS( "Exit");
#endif

  return ( c );
}

UInt8 file_exists ( char *filename ) {
  FILE *f = fopen ( filename, "r" );

  if ( f ) {
    fclose ( f );
    return ( 1 );
  }

  return ( 0 );
}

UInt32 fread ( char *buf, UInt32 size, UInt32 num, FILE *file ) {
  UInt32 result;
  Err error;
  UInt32 ext = size * num;

#ifdef DEBUGFILE
  DEBUGS(__FUNCTION__);
  DEBUGS(file->filename);
  DEBUGU32(num*size);
#endif

  // do we have any data in the buffer already? if so, lets first use
  // that and then pick up any extra needed
  if ( ext < file -> tail ) {
    // buffer already has all the data
    //DEBUGS("less");

    // pick up buffered bytes
    MemMove ( buf, file -> buffer, ext );

    // shuffle buffer up into place
    MemMove ( file -> buffer, file -> buffer + ext, file -> tail - ext );

    // update tail to new size
    file -> tail -= ext;

  } else {
    // buffer doesn't have enough data
    //DEBUGS("more");

    // move buffer into buf
    MemMove ( buf, file -> buffer, file -> tail ); // might be zero

    // go to end of buff
    buf += file -> tail;

    // num remaining
    ext -= file -> tail;

    // empty buffer
    file -> tail = 0; // now empty

    // pick up remaining
    error = VFSFileRead ( file -> ref, ext, buf, &result );

  }

#ifdef DEBUGFILE
  DEBUGS( "Exit");
#endif

  //return ( result );
  return ( size * num );
}
